use memmap2::MmapMut;

use super::{
    error::{MutablePEError, PEParseError},
    header::*,
};
use std::{
    mem::{self, offset_of},
    ops::{Deref, DerefMut, IndexMut},
};

pub trait ReadOnlyPE: Deref<Target = [u8]> + Sized + AsRef<[u8]> {
    /// Get a reference to a type at a given offset.
    fn get_ref<T>(&self, offset: usize) -> Result<&T, PEParseError> {
        let len = std::mem::size_of::<T>();
        if offset + len > self.len() {
            return Err(PEParseError::OutOfBounds);
        }
        let ptr = self.as_ptr().wrapping_offset(offset as isize) as *const T;
        Ok(unsafe { &*ptr })
    }

    // 判断是64位还是32位
    fn is_64_bit(&self) -> Result<bool, PEParseError> {
        // 先以32位加载可选头，通过可选头的Magic字段判断
        let optional_header: &ImageOptionalHeader32 = self.get_optional_header_32()?;
        let is_64_bit = optional_header.magic == 0x20b;
        Ok(is_64_bit)
    }

    /// 将size与节对齐值进行对齐，返回对齐后的值
    /// size: 需要对齐的值
    fn align_size_with_section_alignment(&self, size: u32) -> Result<u32, PEParseError> {
        let nt_header = self.get_nt_header()?;
        let section_alignment = match nt_header {
            ImageNTHeader::NTHeader32(nt_header) => nt_header.optional_header.section_alignment,
            ImageNTHeader::NTHeader64(nt_header) => nt_header.optional_header.section_alignment,
        };
        let aligned_size = (size + section_alignment - 1) & !(section_alignment - 1);
        Ok(aligned_size)
    }

    /// 将size与文件对齐值进行对齐，返回对齐后的值
    /// size: 需要对齐的值
    fn align_size_with_file_alignment(&self, size: u32) -> Result<u32, PEParseError> {
        let nt_header = self.get_nt_header()?;
        let file_alignment = match nt_header {
            ImageNTHeader::NTHeader32(nt_header) => nt_header.optional_header.file_alignment,
            ImageNTHeader::NTHeader64(nt_header) => nt_header.optional_header.file_alignment,
        };
        let aligned_size = (size + file_alignment - 1) & !(file_alignment - 1);
        Ok(aligned_size)
    }

    /// Get the DOS header without verifying its contents.
    fn get_dos_header(&self) -> Result<&ImageDosHeader, PEParseError> {
        let result = self.get_ref::<ImageDosHeader>(0)?;
        Ok(result)
    }

    /// 作为32位的NT头解析
    fn get_nt_header32(&self) -> Result<&ImageNTHeader32, PEParseError> {
        let nt_offset = self.get_nt_headers_offset()?;
        let result = self.get_ref::<ImageNTHeader32>(nt_offset)?;
        Ok(result)
    }
    /// 作为64位的NT头解析
    fn get_nt_header64(&self) -> Result<&ImageNTHeader64, PEParseError> {
        let nt_offset = self.get_nt_headers_offset()?;
        let result = self.get_ref::<ImageNTHeader64>(nt_offset)?;
        Ok(result)
    }
    /// 动态的解析NT头 无论是32位还是64位
    fn get_nt_header(&self) -> Result<ImageNTHeader, PEParseError> {
        let is_64_bit = self.is_64_bit()?;
        let result = match is_64_bit {
            true => ImageNTHeader::NTHeader64(*self.get_nt_header64()?),
            false => ImageNTHeader::NTHeader32(*self.get_nt_header32()?),
        };
        Ok(result)
    }

    // 获取nt头的偏移
    fn get_nt_headers_offset(&self) -> Result<usize, PEParseError> {
        let dos_header = self.get_dos_header()?;
        let nt_offset = dos_header.e_lfanew.0 as usize;
        Ok(nt_offset)
    }

    // 获取文件头的偏移
    fn get_file_header_offset(&self) -> Result<usize, PEParseError> {
        // 1. 获取nt头偏移
        let nt_offset = self.get_nt_headers_offset()?;
        // 2. 计算文件头偏移
        let file_header_offset = nt_offset + std::mem::size_of::<u32>();
        Ok(file_header_offset)
    }

    // 获取文件头数据
    fn get_file_header(&self) -> Result<&ImageFileHeader, PEParseError> {
        let file_header_offset = self.get_file_header_offset()?;
        let result = self.get_ref::<ImageFileHeader>(file_header_offset)?;
        Ok(result)
    }

    // 获取节区数量
    fn get_number_of_sections(&self) -> Result<usize, PEParseError> {
        let file_header = self.get_file_header()?;
        Ok(file_header.number_of_sections as usize)
    }

    // 获取可选头大小
    fn get_size_of_optional_header(&self) -> Result<usize, PEParseError> {
        let file_header = self.get_file_header()?;
        Ok(file_header.size_of_optional_header as usize)
    }
    /// 作为32位的可选头解析
    fn get_optional_header_32(&self) -> Result<&ImageOptionalHeader32, PEParseError> {
        let file_header_offset = self.get_file_header_offset()?;
        let optional_header_offset = file_header_offset + std::mem::size_of::<ImageFileHeader>();
        let result = self.get_ref::<ImageOptionalHeader32>(optional_header_offset)?;
        Ok(result)
    }

    // 作为64位的可选头解析
    fn get_optional_header_64(&self) -> Result<&ImageOptionalHeader64, PEParseError> {
        let file_header_offset = self.get_file_header_offset()?;
        let optional_header_offset = file_header_offset + std::mem::size_of::<ImageFileHeader>();
        let result = self.get_ref::<ImageOptionalHeader64>(optional_header_offset)?;
        Ok(result)
    }
    /// 动态的解析可选头 无论是32位还是64位
    fn get_optional_header(&self) -> Result<ImageOptionalHeader, PEParseError> {
        let is_64_bit = self.is_64_bit()?;
        let result = match is_64_bit {
            true => ImageOptionalHeader::OptionalHeader64(*self.get_optional_header_64()?),
            false => ImageOptionalHeader::OptionalHeader32(*self.get_optional_header_32()?),
        };
        Ok(result)
    }
    /// 获取可选头的偏移
    fn get_optional_header_offset(&self) -> Result<usize, PEParseError> {
        let file_header_offset = self.get_file_header_offset()?;
        let optional_header_offset = file_header_offset + std::mem::size_of::<ImageFileHeader>();
        Ok(optional_header_offset)
    }

    /// 获取节区头的偏移
    fn get_section_headers_offset(&self) -> Result<usize, PEParseError> {
        // 节区头偏移在可选头之后，可选头大小是可变的，所以需要计算
        let optional_header_size = self.get_size_of_optional_header()?;
        let file_header_offset = self.get_file_header_offset()?;
        let section_header_offset =
            file_header_offset + std::mem::size_of::<ImageFileHeader>() + optional_header_size;
        Ok(section_header_offset)
    }

    // 获取所有节区头数据
    fn get_section_headers(&self) -> Result<&[ImageSectionHeader], PEParseError> {
        // 1. 获取节区数量
        let number_of_sections = self.get_number_of_sections()?;
        // 2. 获取节区头偏移
        let section_header_offset = self.get_section_headers_offset()?;
        // 3. 获取节区头数据
        unsafe {
            let ptr = self
                .as_ptr()
                .wrapping_offset(section_header_offset as isize)
                as *const ImageSectionHeader;
            let result = std::slice::from_raw_parts(ptr, number_of_sections);
            Ok(result)
        }
    }

    /// 获取头映射最大可扩容大小
    fn get_header_max_expand_size(&self) -> Result<u32, PEParseError> {
        // 最大可扩容大小 = 节对齐大小 - 最后一个节区头最后一个字节的偏移
        let nt_header: ImageNTHeader = self.get_nt_header()?;
        let (section_alignment, file_alignment) = match nt_header {
            ImageNTHeader::NTHeader32(nt_header) => (nt_header.optional_header.section_alignment, nt_header.optional_header.file_alignment),
            ImageNTHeader::NTHeader64(nt_header) => (nt_header.optional_header.section_alignment, nt_header.optional_header.file_alignment),
        };
        let section_header_offset = self.get_section_headers_offset()?;
        let section_headers = self.get_number_of_sections()? as u32;
        let last_section_header_offset = section_header_offset 
            + (section_headers as usize * std::mem::size_of::<ImageSectionHeader>());
        let mut max_size_of_headers = section_alignment - last_section_header_offset as u32;
        // max_size_of_headers 向下对齐
        max_size_of_headers &= !(file_alignment - 1);

        Ok(max_size_of_headers)
    }


}
/// 可修改的PE trait
pub trait MutablePE: ReadOnlyPE + DerefMut<Target = [u8]> + AsMut<[u8]> {
    /// 实现一个get_mut
    fn get_mut<T>(&mut self, offset: usize) -> Result<&mut T, PEParseError> {
        let len = std::mem::size_of::<T>();
        if offset + len > self.len() {
            return Err(PEParseError::OutOfBounds);
        }
        let ptr = self.as_mut_ptr().wrapping_offset(offset as isize) as *mut T;
        Ok(unsafe { &mut *ptr })
    }

    /// 获取可变的节区头数据
    fn get_section_headers_mut(&mut self) -> Result<&mut [ImageSectionHeader], PEParseError> {
        // 1. 获取节区数量
        let number_of_sections = self.get_number_of_sections()?;
        // 2. 获取节区头偏移
        let section_header_offset = self.get_section_headers_offset()?;
        // 3. 获取节区头数据
        unsafe {
            let ptr = self
                .as_mut_ptr()
                .wrapping_offset(section_header_offset as isize)
                as *mut ImageSectionHeader;
            let result = std::slice::from_raw_parts_mut(ptr, number_of_sections);
            Ok(result)
        }
    }
    /// 获取可变的NT头数据64位
    fn get_nt_header64_mut(&mut self) -> Result<&mut ImageNTHeader64, PEParseError> {
        let nt_offset = self.get_nt_headers_offset()?;
        let result = self.get_mut::<ImageNTHeader64>(nt_offset)?;
        Ok(result)
    }

    /// 获取可变的NT头数据32位
    fn get_nt_header32_mut(&mut self) -> Result<&mut ImageNTHeader32, PEParseError> {
        let nt_offset = self.get_nt_headers_offset()?;
        let result = self.get_mut::<ImageNTHeader32>(nt_offset)?;
        Ok(result)
    }

    /// 获取可变的NT头数据
    fn get_nt_header_mut(&mut self) -> Result<ImageNTHeaderMut, PEParseError> {
        let is_64_bit = self.is_64_bit()?;
        let result = match is_64_bit {
            true => ImageNTHeaderMut::NTHeader64(self.get_nt_header64_mut()?),
            false => ImageNTHeaderMut::NTHeader32(self.get_nt_header32_mut()?),
        };
        Ok(result)
    }

    /// 扩大头映射大小，调用者必须要保证当挪动所有节区数据时，不会超出文件范围
    /// 也就是，调用者必须要在扩大头映射大小之前，先扩大文件大小
    /// add_size: 需要扩大的大小，不需要进行对齐，会自动对齐
    fn expand_headers(&mut self, add_size: u32) -> Result<(), MutablePEError> {
        // aligned_size: 需要增加的大小，对齐后的值
        let aligned_file_size = self.align_size_with_file_alignment(add_size)?;
        let max_expand_size = self.get_header_max_expand_size()?;
        let optional_header = self.get_optional_header()?;
        let origin_size_for_headers = optional_header.get_size_of_headers();
        // 如果扩大后超过了最大可扩容大小，就不允许扩容
        if aligned_file_size > max_expand_size {
            return Err(MutablePEError::CannotExpandHeader(
                aligned_file_size,
                origin_size_for_headers,
            ));
        }
        // 检查一下把，万一调用者没有扩大文件大小
        let sections = self.get_section_headers()?;
        let self_len = self.len();
        let last_section = sections.last().unwrap();
        let last_section_end = last_section.pointer_to_raw_data + last_section.size_of_raw_data;
        if last_section_end as usize + aligned_file_size as usize > self_len {
            return Err(MutablePEError::CannotExpandFileSize(
                last_section_end as u32 + aligned_file_size,
                self_len as u32,
            ));
        }
        let mut nt_header = self.get_nt_header_mut()?;
        let op = nt_header.get_optional_header_mut();

        // 更新可选头的SizeOfHeaders和SizeOfImage
        match op {
            ImageOptionalHeaderMut::OptionalHeader32(op) => {
                op.size_of_headers += aligned_file_size;
            }
            ImageOptionalHeaderMut::OptionalHeader64(op) => {
                op.size_of_headers += aligned_file_size;
            }
        }

        let mut_section = self.get_section_headers_mut()?;

        let start_offset = mut_section.first().unwrap().pointer_to_raw_data;
        let end_offset = last_section_end;
        // 更新节区头的PointerToRawData
        for section in mut_section.iter_mut() {
            section.pointer_to_raw_data += aligned_file_size;
        }
        {
            let self_data = self.as_mut();
            // 把[start_offset, end_offset)的数据往后挪动aligned_file_size
            self_data.copy_within(
                start_offset as usize..end_offset as usize,
                (start_offset + aligned_file_size) as usize,
            );
        }
        Ok(())
    }
}
